from enum import Enum


class Color(Enum):
    """
    Light or surface color.
    """
    BLACK = ...
    BLUE = ...
    GREEN = ...
    YELLOW = ...
    RED = ...
    WHITE = ...
    BROWN = ...
    ORANGE = ...
    PURPLE = ...
