from pybricks.parameters import Button
from typing import List


class Buttons:
    """
    A stub class to represent the buttons member of the EV3Brick class.
    """

    def __init__(self):
        ...

    def pressed(self) -> List[Button]:
        """
        Checks which buttons are currently pressed.

        Returns:
        List of pressed buttons.
        """
        return []
